# (c) Copyright 2019 Trent Hauck
# All Rights Reserved
"""The base tokenizer."""

from typing import Optional, List

from pydantic import BaseSettings, Field, root_validator


class SequenceTokenizerSettings(BaseSettings):
    """The base tokenizer settings."""

    bos_token: Optional[str] = Field(None, env="GCGC_BOS_TOKEN")
    eos_token: Optional[str] = Field(None, env="GCGC_EOS_TOKEN")
    unk_token: Optional[str] = Field("?", env="GCGC_UNK_TOKEN")
    mask_token: Optional[str] = Field(None, env="GCGC_MASK_TOKEN")
    pad_token: Optional[str] = Field(None, env="GCGC_PAD_TOKEN")

    max_length: Optional[int] = Field(None, env="GCGC_MAX_LENGTH")
    min_length: Optional[int] = Field(None, env="GCGC_MIN_LENGTH")
    conform_length: Optional[int] = Field(None, env="GCGC_CONFORM_LENGTH")

    @root_validator
    def validate_lengths(cls, values):  # pylint: disable=no-self-argument, no-self-use
        """Check the length arguments are valid."""
        max_length = values.get("max_length")
        min_length = values.get("min_length")
        conform_length = values.get("conform_length")
        pad_token = values.get("pad_token")

        if conform_length is not None and (max_length is not None or min_length is not None):
            raise ValueError(f"If conform length is not None, max and min length can't be set.")

        if max_length is not None and min_length is not None and min_length > max_length:
            raise ValueError(
                f"Min length {min_length} cannot be more than max length {max_length}."
            )

        if (conform_length is not None or min_length is not None) and pad_token is None:
            raise ValueError("Cannot set conform_length or min_length without a pad_token.")

        return values


def _pad_token_list(tokens: List[str], pad_to: int, pad_char: str):
    """Pad the input tokens list to tokens."""
    token_len = len(tokens)

    if token_len > pad_to:
        return tokens

    new_tokens = pad_to - token_len
    return tokens + [pad_char] * new_tokens


class SequenceTokenizer:
    """The sequence tokenizer object."""

    def __init__(self, settings: Optional[SequenceTokenizerSettings] = None):
        """Inits the sequence tokenizer with a set of settings."""
        self.settings = settings or SequenceTokenizerSettings()

    def apply_length_constraints(self, tokens: List[str]):
        """Apply the constraints of the settings to the passed tokens list."""
        if self.settings.conform_length:
            tokens = _pad_token_list(tokens, self.settings.conform_length, self.settings.pad_token)
            return tokens[: self.settings.conform_length]

        if self.settings.min_length:
            tokens = _pad_token_list(tokens, self.settings.min_length, self.settings.pad_token)

        if self.settings.max_length:
            tokens = tokens[: self.settings.max_length]

        return tokens



# (c) Copyright 2019 Trent Hauck
# All Rights Reserved
"""Module for KMER tokenization."""

from typing import List, Optional
import itertools as it

from pydantic import Field, validator

from gcgc.tokenizer.base import SequenceTokenizer, SequenceTokenizerSettings
from gcgc import alphabets


class KmerTokenizerSettings(SequenceTokenizerSettings):
    """The specification for the tokenizer."""

    alphabet: str = Field("ATCG", env="GCGC_ALPHABET")
    kmer_length: int = Field(1, env="GCGC_KMER_LENGTH")
    kmer_stride: int = Field(1, env="GCGC_KMER_STRIDE")

    @validator("alphabet")
    def resolve_alphabet(cls, alphabet):  # pylint: disable=no-self-use, no-self-argument
        """Resolve the alphabet if it's a named alphabet."""
        return alphabets.resolve_alphabet(alphabet)

    @property
    def special_tokens(self) -> List[str]:
        """Return the list of special tokens that are not None."""
        return [
            x
            for x in [
                self.bos_token,
                self.eos_token,
                self.unk_token,
                self.pad_token,
                self.mask_token,
            ]
            if x is not None
        ]


def _create_kmer_vocab_from_token(
    alphabet, kmer_length, token_characters: Optional[List[str]] = None
):
    """Create vocabulary object from a list of tokens."""
    if token_characters is None:
        token_characters = []

    token_to_int = {}

    for i, token in enumerate(token_characters):
        token_to_int[token] = i

    token_list = ["".join(kmer) for kmer in it.product(list(alphabet), repeat=kmer_length)]
    for i, token in enumerate(token_list, start=len(token_characters)):
        token_to_int[token] = i

    # pylint: disable=too-many-function-args
    return token_to_int


class KmerTokenizer(SequenceTokenizer):
    """A sequence tokenizer."""

    def __init__(self, settings: Optional[KmerTokenizerSettings] = None):
        """Init the SequenceTokenizer class.

        Args:
            settings: The settings for the tokenizer.
            vocabulary: The vocabulary for the tokenizer.

        """
        self.settings = settings or KmerTokenizerSettings()
        super().__init__(settings)

        self.vocab = _create_kmer_vocab_from_token(
            self.settings.alphabet, self.settings.kmer_length, self.settings.special_tokens
        )

    def _kmer_n(self, seq: str) -> List[str]:
        seq_len = len(seq)
        iterations = seq_len - self.settings.kmer_length + 1
        kmers = []

        for i in range(0, iterations, self.settings.kmer_stride):
            kmer = seq[i : i + self.settings.kmer_length]
            kmers.append(kmer)

        return kmers

    def encode(self, seq: str) -> List[int]:
        """Encode the underlying sequence into a list of tokens."""
        return [
            self.vocab.get(s, self.vocab.get(self.settings.unk_token))
            for s in self.encode_as_tokens(seq)
        ]

    def encode_as_tokens(self, seq: str) -> List[str]:
        """Tokenize the sequence into a list of token tokens.

        Args:
            seq: The sequence to encode.

        Returns:
            The list of strs that are the tokens.

        """
        seq_len = len(seq)

        if seq_len < self.settings.kmer_length:
            raise ValueError(
                f"seq length {seq_len} cannot be less than the kmer "
                "size {self.settings.kmer_length}"
            )

        if self.settings.kmer_length == 1:
            kmer_list = list(seq)
        else:
            kmer_list = self._kmer_n(seq)

        if self.settings.bos_token:
            kmer_list = [self.settings.bos_token] + kmer_list

        if self.settings.eos_token:
            kmer_list = kmer_list + [self.settings.eos_token]

        return super().apply_length_constraints(kmer_list)



# (c) Copyright 2019 Trent Hauck
# All Rights Reserved
"""Stores the alphabets, these are the same as BioPython."""

_ALPHABETS = {
    "protein": "ACDEFGHIKLMNPQRSTVWY",
    "extended_protein": "ACDEFGHIKLMNPQRSTVWYBXZJUO",
    "ambiguous_dna": "GATCRYWSMKHBVDN",
    "unambiguous_dna": "GATC",
    "extended_dna": "GATCBDSW",
    "ambiguous_rna": "GAUCRYWSMKHBVDN",
    "unambiguous_rna": "GAUC",
}


def resolve_alphabet(alphabet: str) -> str:
    """Try to get the alphabet from the known list, otherwise, return what was passed.

    Args:
        alphabet: The alphabet to resolve.

    Returns:
        The resolved alphabet.

    """
    try:
        return _ALPHABETS[alphabet]
    except KeyError:
        return alphabet



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Test the alphabet."""

import pytest

from gcgc.tokenizer import KmerTokenizer, KmerTokenizerSettings


@pytest.mark.parametrize(
    "seq,settings,expected_tokens,expected_encoding",
    [
        (
            "ATCG",
            KmerTokenizerSettings(alphabet="ATCG", kmer_length=1, kmer_stride=1),
            list("ATCG"),
            [1, 2, 3, 4],
        ),
        (
            "ATCG",
            KmerTokenizerSettings(alphabet="ATCG", max_length=2, kmer_length=1, kmer_stride=1),
            list("AT"),
            [1, 2],
        ),
        (
            "ATCG",
            KmerTokenizerSettings(
                alphabet="ATCG", min_length=4, pad_token="|", kmer_length=2, kmer_stride=1
            ),
            ["AT", "TC", "CG", "|"],
            [3, 8, 13, 1],
        ),
        (
            "AATT",
            KmerTokenizerSettings(
                alphabet="ATCG", pad_token="|", conform_length=5, kmer_length=2, kmer_stride=2
            ),
            ["AA", "TT", "|", "|", "|"],
            [2, 7, 1, 1, 1],
        ),
        (
            "ATCGATCGATCGATCG",
            KmerTokenizerSettings(alphabet="ATCG", max_length=2, kmer_length=2, kmer_stride=1),
            ["AT", "TC"],
            [2, 7],
        ),
        (
            "ATCGATCG",
            KmerTokenizerSettings(
                alphabet="ATCG",
                bos_token=">",
                eos_token="<",
                max_length=10,
                kmer_length=2,
                kmer_stride=1,
            ),
            [">", "AT", "TC", "CG", "GA", "AT", "TC", "CG", "<"],
            [0, 4, 9, 14, 15, 4, 9, 14, 1],
        ),
    ],
)
def test_kmer_tokenization(seq, settings, expected_tokens, expected_encoding):
    """Test the kemrs are encoded as expected."""
    tokenizer = KmerTokenizer(settings)
    actual_tokens = tokenizer.encode_as_tokens(seq)
    assert actual_tokens == expected_tokens

    actual_encoded = tokenizer.encode(seq)
    assert actual_encoded == expected_encoding



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Top-level GCGC module."""

import warnings as _warnings

from gcgc.cli import cli

__version__ = "0.12.0-dev.7"

_warnings.simplefilter(action="ignore", category=PendingDeprecationWarning)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Test the sentence models."""

import numpy as np
from gcgc.tokenizer import sentence_piece_tokenizer


def test_train_sentence_piece(tmp_path):
    """Test fitting the sentence piece model."""
    sequences = ["".join(x) for x in np.random.choice(list("ATCG"), (20, 20))]

    text_file = tmp_path / "sp.txt"
    text_file.write_text("\n".join(sequences))

    model_prefix = tmp_path / "model"

    settings = sentence_piece_tokenizer.BioSequencePieceSettings(
        model_prefix=model_prefix, vocab_size=20, bos_token=">", eos_token="<", pad_token="|"
    )
    sp_tokenizer = sentence_piece_tokenizer.BioSequencePiece(settings)
    sp_tokenizer.fit_on_text(text_file)

    tokenized = sp_tokenizer.encode_as_tokens("ATCGATCGATCG")
    assert all([isinstance(x, str) for x in tokenized])

    ids = sp_tokenizer.encode("ATCGATCGATCG")
    assert all([isinstance(x, int) for x in ids])



# (c) Copyright 2019 Trent Hauck
# All Rights Reserved
"""Module for Sentence Piece tokenization."""

import tempfile
from typing import List, Optional, Dict
from pathlib import Path
import shutil

from pydantic import Field
from typing_extensions import Literal

from Bio import SeqIO
from gcgc.tokenizer.base import SequenceTokenizer, SequenceTokenizerSettings

try:
    import sentencepiece as spm

    # pylint: disable=invalid-name
    has_spm = True
except ImportError:
    # pylint: disable=invalid-name
    has_spm = False


class BioSequencePieceSettings(SequenceTokenizerSettings):
    """The settings for the sentence piece model."""

    model_prefix: Path = Field(..., env="GCGC_SP_MODEL_PREFIX")
    vocab_size: int = Field(8000, env="GCGC_SP_VOCAB_SIZE")
    model_type: Literal["unigram", "bpe"] = "unigram"
    max_sequence_length: int = 4192

    @property
    def model_path(self) -> Path:
        """Return the model path based on the prefix."""
        return self.model_prefix.with_suffix(".model")

    @property
    def model_vocab(self) -> Path:
        """Return the model vocab based on the prefix."""
        return self.model_prefix.with_suffix(".vocab")


class BioSequencePiece(SequenceTokenizer):
    """A sentence piece for model on biological sequences."""

    def __init__(self, settings: Optional[BioSequencePieceSettings] = None):
        """Init the BioSequencePiece class.

        Args:
            settings: The settings for the tokenizer.

        """
        if not has_spm or not shutil.which("spm_train"):
            raise RuntimeError("Trying to use sentencepiece but the python library is missing!")

        self.settings = settings or BioSequencePieceSettings()

        self.vocab: Dict[str, int] = {}
        self._sp_processor = None

    @property
    def sp_processor(self):
        """Return the SequencePiece process object."""
        if self._sp_processor is not None:
            return self._sp_processor

        self._sp_processor = spm.SentencePieceProcessor()
        self._sp_processor.load(str(self.settings.model_path))
        return self._sp_processor

    def fit_on_fasta(self, fasta_file: Path):
        """Run the the SP algo on the fasta_file."""
        with tempfile.TemporaryDirectory() as tmpdir:
            tmppath = Path(tmpdir)

            text_file_path = tmppath / "input_textfiles.txt"
            with text_file_path.open("w") as text_lines, fasta_file.open("r") as input_handler:
                for record in SeqIO.parse(input_handler, "fasta"):
                    text_lines.write(f"{str(record.seq)}\n")

            self.fit_on_text(text_file_path)

    def fit_on_text(self, text_file: Path):
        """Run the the SP algo on the text_file."""
        args = [
            f"--input={str(text_file)}",
            f"--model_prefix={self.settings.model_prefix}",
            f"--vocab_size={self.settings.vocab_size}",
            f"--model_type={self.settings.model_type}",
            f"--max_sentence_length={self.settings.max_sequence_length}",
        ]

        if self.settings.unk_token:
            args.extend([f"--unk_piece={self.settings.unk_token}"])
        else:
            args.extend(["--unk_id=-1"])

        if self.settings.bos_token:
            args.extend([f"--bos_piece={self.settings.bos_token}"])
        else:
            args.extend(["--bos_id=-1"])

        if self.settings.eos_token:
            args.extend([f"--eos_piece={self.settings.eos_token}"])
        else:
            args.extend(["--eos_id=-1"])

        if self.settings.pad_token:
            args.extend([f"--pad_piece={self.settings.pad_token}", "--pad_id=-1"])
            self.vocab[self.settings.pad_token] = -1
        else:
            args.extend(["--pad_id=-1"])

        spm.SentencePieceTrainer.Train(" ".join(args))

        self.load_vocab()

    def encode(self, seq: str) -> List[int]:
        """Encode the underlying sequence into a list of tokens."""
        return [
            self.vocab.get(s, self.vocab.get(self.settings.unk_token))
            for s in self.encode_as_tokens(seq)
        ]

    def encode_as_tokens(self, seq: str) -> List[str]:
        """Tokenize the sequence into a list of token tokens.

        Args:
            seq: The sequence to encode.

        Returns:
            The list of strs that are the tokens.

        """
        return super().apply_length_constraints(self.sp_processor.EncodeAsPieces(seq))

    def load_vocab(self):
        """Load the vocabulary from the file."""
        for line, token in enumerate(self.settings.model_vocab.open()):
            token = token.strip("\n").split("\t")[0]
            self.vocab[token] = line



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Contains tokenizer level packages."""

from gcgc.tokenizer.kmer_tokenzier import KmerTokenizer
from gcgc.tokenizer.kmer_tokenzier import KmerTokenizerSettings

__all__ = ["KmerTokenizer", "KmerTokenizerSettings"]
